const trailingNumbersRegExp = /\d+$/g;
const stripTrailingNumbers = (x) => x.replace(trailingNumbersRegExp, '');
const thru = (x) => x;
const createContext = (extraTypes = [], ptrs = new Map()) => {
    const Types = new Map([
        [Object],
        [Array, [thru, thru]],
        [Map, [x => [...x]]],
        [Set, [x => [...x]]],
        [Date],
        ...extraTypes,
    ]);
    const TypesMap = [...Types.keys()].map(x => [stripTrailingNumbers(x.name), x]);
    let pointer = 0;
    const defaultSerializer = (obj) => obj.toJSON ? obj.toJSON() : Object.assign({}, obj);
    const createEntry = (serializer, obj) => ptrs.get(obj) ?? ({
        $r: (ptrs.set(obj, { $p: ++pointer }), pointer),
        $t: stripTrailingNumbers(obj.constructor.name),
        $v: serializer(obj),
    });
    const replacer = (top, clear = true) => (clear && (pointer = 0, ptrs.clear()), function (key) {
        const value = this[key];
        if ((value === top && key !== '')
            || (typeof value === 'object' && value.$r)
            || this.$r) {
            return value;
        }
        if (typeof value === 'object') {
            const [serializer = defaultSerializer] = Types.get(value.constructor) ?? [];
            return createEntry(serializer, value);
        }
        return value;
    });
    const reviver = (classes, refs = new Map(), pending = new Map()) => {
        const types = new Map([
            ...TypesMap,
            ...classes.map(x => [stripTrailingNumbers(x.name), x]),
        ]);
        const getRef = (value, key, owner) => {
            const { $p } = value;
            if (refs.has($p))
                return refs.get($p);
            else {
                let queued = pending.get($p);
                if (!queued)
                    pending.set($p, queued = new Set());
                queued.add([owner, key]);
                return value;
            }
        };
        return function (key, value) {
            if (typeof value === 'object' && value !== null) {
                if (value.$t) {
                    const { $r, $t, $v } = value;
                    const ctor = types.get($t);
                    if (!ctor) {
                        throw new TypeError('Unable to deserialize object type: ' + $t
                            + `\n  Be sure to pass the class like so:\n\n    deserialize(serialized, [${$t}])`);
                    }
                    const [, deserializer = ((x) => new ctor(x))] = Types.get(ctor) ?? [];
                    const result = deserializer($v);
                    if ($t === 'Map') {
                        for (const pv of pending.values()) {
                            for (const q of pv) {
                                if ($v.includes(q[0])) {
                                    if (q[1] == '0') {
                                        q[1] = (key) => {
                                            const p = q[0][0];
                                            const v = result.get(p);
                                            result.set(key, v);
                                            result.delete(p);
                                        };
                                    }
                                    else if (q[1] == '1') {
                                        q[1] = (value) => {
                                            let p = q[0][0];
                                            if (p?.$p) {
                                                p = refs.get(p.$p);
                                            }
                                            result.set(p, value);
                                        };
                                    }
                                }
                            }
                        }
                    }
                    if ($t === 'Set') {
                        for (const pv of pending.values()) {
                            for (const q of pv) {
                                if ($v.includes(q[0][0])) {
                                    const p = q[0][q[1]];
                                    q[1] = (value) => {
                                        result.add(value);
                                        result.delete(p);
                                    };
                                }
                            }
                        }
                    }
                    refs.set($r, result);
                    if (pending.has($r)) {
                        pending.get($r).forEach(([parent, key]) => {
                            if (typeof key === 'function') {
                                key(result);
                            }
                            else {
                                parent[key] = result;
                            }
                        });
                    }
                    return result;
                }
                else if (value.$p) {
                    return getRef(value, key, this);
                }
            }
            return value;
        };
    };
    return { replacer, reviver };
};
const { replacer, reviver } = createContext();
const serialize = (any, indent, clear) => JSON.stringify(any, replacer(any, clear), indent);
const deserialize = (serialized, classes = []) => JSON.parse(serialized, reviver(classes));

export { createContext, deserialize, replacer, reviver, serialize, thru };
//# sourceMappingURL=serialize-whatever.js.map
